/* tfontedpr - general purpose "pretty printer" for use with TeX.
 *
 * Copyright (C) 1985 by Van Jacobson, Lawrence Berkeley Laboratory.
 * This program may be freely used and copied but may not be sold
 * without the author's written permission.  This notice must remain
 * in any copy or derivative.
 *
 * This program is used as part of the "tgrind" shell script.  It
 * converts program source file(s) to TeX input files.
 *
 * This program is an adaptation of "vfontedpr" v4.2 (12/11/84) from
 * the 4.2bsd Unix distribution.  Vfontedpr was written by Dave
 * Presotto (based on an earlier program of the same name written by
 * Bill Joy).
 *
 * I would welcome comments, enhancements, bug fixes, etc.  Please
 * mail them to:
 *	van@lbl-rtsg.arpa	(from arpanet, milnet, csnet, etc.)
 *	..!ucbvax!lbl-csam!van	(from Usenet/UUCP)
 *
 * Modifications.
 * --------------
 * 30Mar85  Chris & Van: Fixed "\C" & "\S" (comment & string start indicators)
 *			to really appear at the start of comments & strings.
 *			Changes for speeded-up expmatch.
 * 29Mar85  Chris Torek (chris@maryland):  Bug fixes for '~' and '^L'
 *			output.  Most cpu-time eaters recoded to improve
 *			efficiency.
 * 10Feb85  Van		Written.
 */

#include <config.h>			/* private file */

#if defined(HAVE_CTYPE_H)
#include <ctype.h>
#endif

#if defined(HAVE_STDIO_H)
#include <stdio.h>
#endif

#if defined(HAVE_STRING_H)
#include <string.h>
#endif

#if defined(HAVE_SYS_TYPES_H)
#include <sys/types.h>
#endif

#if defined(HAVE_SYS_STAT_H)
#include <sys/stat.h>
#endif

#if defined(HAVE_STDLIB_H)
#include <stdlib.h>
#endif

#if defined(HAVE_TIME_H)
#include <time.h>
#else
extern char	*ctime ARGS((const time_t *));
#endif

#include <stdc.h>			/* private file */
#include <boolean.h>			/* private file */
#include <bufsiz.h>			/* private file */
#include <regexp.h>			/* private file */
#include <vgrindefs.h>			/* private file */

#define CHRVAL(c)	(0xff & (c))
#define elementsof(v)	(sizeof(v)/sizeof(v[0]))
#define NIL		((char*)NULL)
#define NUL		'\0'
#define STANDARD	0
#define ALTERNATE	1

#define STRLEN		10	/* length of strings introducing things */
#define PNAMELEN 	80	/* length of a function/procedure name */
#define PSMAX 		20	/* size of procedure name stacking */


/* regular expression routines */

static int		iskw ARGS((const char *s_));
static boolean		isproc ARGS((const char *s_));
int			main ARGS((int argc_, register char **argv_));
static void		putKcp ARGS((const char *start_, const char *end_,
				     boolean nix_));
static void		putScp ARGS((const char *os_));
static void		putstr ARGS((const char *cp_));
static int		width ARGS((const char *s_, const char *os_));

#if !defined(LINEAR_KEYWORD_SEARCH)
static unsigned int	Hash ARGS((const char *key_, int len_));
static void		Hash_install ARGS((const char *key_, int len_));
static boolean		Hash_lookup ARGS((const char *key_, int len_));
static unsigned int	Hash_used = 0;
#endif

/*
 *	The state variables
 */

static boolean	incomm;			/* in a comment of the primary type */
static boolean	instr;			/* in a string constant */
static boolean	inchr;			/* in a string constant */
static boolean	nokeyw = FALSE;		/* no keywords being flagged */
static boolean	prccont;		/* continue last procedure */
static int	comtype;		/* type of comment */
static int	psptr;		/* the stack index of the current procedure */
static char	pstack[PSMAX][PNAMELEN+1];	/* the procedure name stack */
static int	plstack[PSMAX];		/* the procedure nesting level stack */
static int	blklevel;		/* current nesting level */
static const char *defsfile = DEFSFILE;	/* name of language definitions file */
static char	pname[MAXBUF+1];

/*
 *	The language specific globals
 */

static const char *language = "c";	/* the language indicator */
#if defined(LINEAR_KEYWORD_SEARCH)
static char	*l_keywds[MAXBUF/2];	/* keyword table address */
#else
static const char *l_keywds[MAXHASH];	/* keyword table address */
#endif
static char	*l_prcbeg;		/* regular expr for procedure begin */
static char	*l_combeg;		/* string introducing a comment */
static char	*l_comend;		/* string ending a comment */
static char	*l_acmbeg;		/* string introducing a comment */
static char	*l_acmend;		/* string ending a comment */
static char	*l_blkbeg;		/* string begining of a block */
static char	*l_blkend;		/* string ending a block */
static char    *l_strbeg;		/* delimiter for string constant */
static char    *l_strend;		/* delimiter for string constant */
static char    *l_chrbeg;		/* delimiter for character constant */
static char    *l_chrend;		/* delimiter for character constant */

#if 0
static char	l_escape;		/* character used to escape characters */
#endif

static boolean	l_toplex;		/* procedures only defined at top lex level */
static boolean	l_onecase;		/* upper & lower case equivalent */


/*
 * The following table converts ASCII characters to a printed
 * representation, taking care of all the TeX quoting.  N.B.: all
 * single-character strings are assumed to be equivalent to the
 * character for that index (i.e., printtab['c'] can't be "f").
 * (This is purely for efficiency hacking.)
 */
const char *printtab[128] = {
    "\0x",   "\\^A",  "\\^B",  "\\^C",  "\\^D",  "\\^E",  "\\^F",  "\\^G",
    "\\^H",  "\t",    "}}\n",  "\\^K",  "\0x",   "\\^M",  "\\^N",  "\\^O",
    "\\^P",  "\\^Q",  "\\^R",  "\\^S",  "\\^T",  "\\^U",  "\\^V",  "\\^W",
    "\\^X",  "\\^Y",  "\\^Z",  "\\^[",  "\\^\\!","\\^]",  "\\^\\^","\\^_",
    " ",     "!",     "\\\"",  "\\#",   "\\$",   "\\%",   "\\&",   "\\'",
    "(",     ")",     "*",     "+",     ",",     "\\-",   ".",     "\\/",
    "0",     "1",     "2",     "3",     "4",     "5",     "6",     "7",
    "8",     "9",     ":",     ";",     "\\<",   "=",     "\\>",   "?",
    "@",     "A",     "B",     "C",     "D",     "E",     "F",     "G",
    "H",     "I",     "J",     "K",     "L",     "M",     "N",     "O",
    "P",     "Q",     "R",     "S",     "T",     "U",     "V",     "W",
    "X",     "Y",     "Z",     "[",     "\\!",   "]",     "\\^",   "\\_",
    "`",     "a",     "b",     "c",     "d",     "e",     "f",     "g",
    "h",     "i",     "j",     "k",     "l",     "m",     "n",     "o",
    "p",     "q",     "r",     "s",     "t",     "u",     "v",     "w",
    "x",     "y",     "z",     "\\{",   "\\|",   "\\}",   "\\~",   "\\^?",
};

/*
 * Output a TeX command to tab to column "col" (see tgrindmac.tex for a
 * partial explanation of the bizarre brace arrangement).
 */
#define tabto(col) (void)printf("}\\Tab{%d}{", col);

#if STDC
int
main(int argc, register char **argv)
#else
int
main(argc, argv)
    int argc;
    register char *argv[];
#endif
{
    const char *fname = "";
    char *p;
    struct stat stbuf;
    char buf[MAXBUF];
    char strings[2 * MAXBUF];
    char defs[2 * MAXBUF];

    (void)printf("\\ifx \\BeginTgrind \\undefined\t%% if macros loaded already, then save time\n");
    (void)printf("\t\\input tgrindmac\n");
    (void)printf("\\fi\n");
    (void)printf("\\BeginTgrind\n");
    argc--, argv++;
    do {
	register char *cp;
	register int i;

	if (argc > 0) {
	    if (!strcmp(argv[0], "-h")) {
		if (argc == 1) {
		    (void)printf("\\Head{}\n");
		    argc = 0;
		    goto rest;
		}
		(void)printf("\\Head{");
		putstr( argv[1] );
		(void)printf( "}\n" );
		argc--, argv++;
		argc--, argv++;
		if (argc > 0)
		    continue;
		goto rest;
	    }

	    /* take input from the standard place */
	    if (!strcmp(argv[0], "-")) {
		argc = 0;
		goto rest;
	    }

	    /* indicate no keywords */
	    if (!strcmp(argv[0], "-n")) {
		nokeyw = TRUE;
		argc--, argv++;
		continue;
	    }

	    /* specify the language */
	    if (!strncmp(argv[0], "-l", 2)) {
	        if (argv[0][2] == '\0')	/* have -l LANGUAGE */
		{
		    argc--, argv++;		  
		    language = argv[0];
		    argc--, argv++;
		}
		else			/* have -lLANGUAGE */
		{
		    language = argv[0]+2;
		    argc--, argv++;
		}
		continue;
	    }

	    /* specify the language description file */
	    if (!strncmp(argv[0], "-d", 2)) {
		defsfile = argv[1];
		argc--, argv++;
		argc--, argv++;
		continue;
	    }

	    /* open the file for input */
	    if (freopen(argv[0], "r", stdin) == NULL) {
		perror(argv[0]);
		exit(EXIT_FAILURE);
	    }

	    fname = argv[0];
	    argc--, argv++;
	}
    rest:

	/*
	 *  get the  language definition from the defs file
	 */
	i = tgetent (defs, language, defsfile);
	if (i == 0) {
	    (void)fprintf (stderr, "no entry for language %s\n", language);
	    exit (EXIT_FAILURE);
	} else  if (i < 0) {
	    (void)fprintf (stderr,  "cannot find vgrindefs file %s\n", defsfile);
	    exit (EXIT_FAILURE);
	}
	l_onecase = tgetflag ("oc") ? TRUE : FALSE;
	if ( l_onecase )
	    re_strncmp = lc_strncmp;
	else
	    re_strncmp = strncmp;
	p = strings;
	if (tgetstr ("kw", &p) == NIL)
	    nokeyw = TRUE;
	else  {
	    if (l_onecase)
	    {	/* re_strncmp() lowercases only its first argument, so */
		/* ensure that the keyword table is stored in lowercase. */
		for (cp = strings; *cp; ++cp)
		{
		    if (isupper(*cp))
			*cp = tolower(*cp);
		}
	    }
	    {
#if defined(LINEAR_KEYWORD_SEARCH)
		/* Old slow linear search keyword table */
		char **cpp;

		cpp = l_keywds;
		cp = strings;
		while (*cp)
		{
		    while (*cp == ' ' || *cp =='\t')
			*cp++ = NUL;
		    if (*cp)
			*cpp++ = cp;
		    while (*cp != ' ' && *cp  != '\t' && *cp)
			cp++;
		}
		*cpp = NIL;
#else
		/* New fast hash search keyword table */
		int k;

		for (k = 0; k < MAXHASH; ++k) /* clear hash table */
		    l_keywds[k] = (const char*)NULL;

		cp = strings;
		while (*cp)
		{
		    char *keystart;

		    while (*cp == ' ' || *cp =='\t')
			*cp++ = NUL;
		    keystart = cp;
		    while (*cp != ' ' && *cp  != '\t' && *cp)
			cp++;
		    Hash_install(keystart, (int)(cp - keystart));
		}
#endif
	    }
	}
	p = buf;
	l_prcbeg = convexp (tgetstr ("pb", &p));
	p = buf;
	l_combeg = convexp (tgetstr ("cb", &p));
	p = buf;
	l_comend = convexp (tgetstr ("ce", &p));
	p = buf;
	l_acmbeg = convexp (tgetstr ("ab", &p));
	p = buf;
	l_acmend = convexp (tgetstr ("ae", &p));
	p = buf;
	l_strbeg = convexp (tgetstr ("sb", &p));
	p = buf;
	l_strend = convexp (tgetstr ("se", &p));
	p = buf;
	l_blkbeg = convexp (tgetstr ("bb", &p));
	p = buf;
	l_blkend = convexp (tgetstr ("be", &p));
	p = buf;
	l_chrbeg = convexp (tgetstr ("lb", &p));
	p = buf;
	l_chrend = convexp (tgetstr ("le", &p));

	p = buf;
	set_isidchar(tgetstr ("id", &p));

	p = buf;
	unset_isidchar(tgetstr ("ni", &p));

	/* Initial characters must be set last, because they default to
	   the idchar set if not supplied. */
	p = buf;
	set_isinichar(tgetstr ("ic", &p));

	p = buf;
	unset_isinichar(tgetstr ("nc", &p));

#if 0
	l_escape = '\\';
#endif

	l_toplex = tgetflag ("tl") ? TRUE : FALSE;

	/* initialize the program */

	incomm = FALSE;
	instr = FALSE;
	inchr = FALSE;
	_escaped = FALSE;
	blklevel = 0;
	for (psptr=0; psptr<PSMAX; psptr++) {
	    pstack[psptr][0] = NUL;
	    plstack[psptr] = 0;
	}
	psptr = -1;
	fstat(fileno(stdin), &stbuf);
	cp = ctime(&stbuf.st_mtime);
	cp[10] = NUL;
	cp[16] = NUL;
	cp[24] = NUL;
	(void)printf("\\TgrindFile{");
	putstr( fname );
	(void)printf("}{%s}{%s %s}\n", cp+11, cp+4, cp+20);

	/*
	 *	MAIN LOOP!!!
	 */
	while (fgets(buf, sizeof buf, stdin) != NULL) {
	    cp = buf;
	    if (*cp == '\f') {
		(void)printf("\\TgrindNewPage\n");
		cp++;
		if (*cp == '\n')/* some people like ^Ls on their own line */
		    continue;
	    }
	    prccont = FALSE;
	    (void)printf("\\L{\\LB{");
	    putScp(cp);
	    if (prccont && (psptr >= 0)) {
	    	(void)printf("\\ProcCont{");
		putstr( pstack[psptr] );
		(void)printf("}\n");
	    }
#ifdef DEBUG
	    (void)printf ("%%%%DEBUG: com %o str %o chr %o ptr %d\n", incomm, instr, inchr, psptr);
#endif
	}
    } while (argc > 0);
    (void)printf("\\EndTgrind\n");
    return(0);
}


#if STDC
void
outchar(unsigned int c)
#else
void
outchar(c)
unsigned int c;
#endif
{
    c = CHRVAL(c);
    if ((c < elementsof(printtab)) && printtab[c][1])
	(void)printf("%s", printtab[c]);
    else
	putchar(c);
}


#if 0
#define isidchr(c) (isalnum(c) || (c) == '_')
#define isinichr(c) isidchr(c)
#else
#define isidchr(c) isidchar[CHRVAL(c)]
#define isinichr(c) isinichar[CHRVAL(c)]
#endif

#if STDC
static void
putScp(const char *os)
#else
static void
putScp(os)
    const char *os;
#endif
{
    register const char *s = os;	/* pointer to unmatched string */
    const char *comptr;			/* start of a comment delimiter */
    const char *comendptr;		/* end of a comment delimiter */
    const char *acmptr;			/* start of an alt. comment delimiter */
    const char *acmendptr;		/* end of an alt. comment delimiter */
    const char *strptr;			/* start of a string delimiter */
    const char *strendptr;		/* end of a string delimiter */
    const char *chrptr;			/* start of a char. const delimiter */
    const char *chrendptr;		/* end of a char. const delimiter */
    const char *blksptr;		/* start of a lexical block start */
    const char *blksendptr;		/* end of a lexical block start */
    const char *blkeptr;		/* start of a lexical block end */
    const char *blkeendptr;		/* end of a lexical block end */

    Str_Start = os;			/* remember the start for expmatch */
    _escaped = FALSE;
    if (nokeyw || incomm || instr)
	goto skip;
    if (isproc(s)) {
	(void)printf("\\Proc{");
	putstr(pname);
	(void)printf("}");
	/* Special hack: C has no nested procedures, but with K&R and
	ANSI/ISO Standard C code in one file, there may be multiple
	procedure headers inside prepreprocessor conditionals. */
	if (strcmp(language,"c") == 0)
	    goto skip;
	if (psptr < (PSMAX - 1)) {
	    ++psptr;
	    strncpy (pstack[psptr], pname, PNAMELEN);
	    pstack[psptr][PNAMELEN] = NUL;
	    plstack[psptr] = blklevel;
	}
    }
skip:
    do {
	/* check for string, comment, blockstart, etc */
	if (!incomm && !instr && !inchr) {

	    blkeendptr = expmatch (s, l_blkend, &blkeptr, NIL);
	    blksendptr = expmatch (s, l_blkbeg, &blksptr, NIL);
	    comendptr = expmatch (s, l_combeg, &comptr, NIL);
	    acmendptr = expmatch (s, l_acmbeg, &acmptr, NIL);
	    strendptr = expmatch (s, l_strbeg, &strptr, NIL);
	    chrendptr = expmatch (s, l_chrbeg, &chrptr, NIL);

	    /* start of a comment? */
	    if (comptr != NIL
		  && (strptr  == NIL || comptr < strptr)
		  && (acmptr  == NIL || comptr < acmptr)
		  && (chrptr  == NIL || comptr < chrptr)
		  && (blksptr == NIL || comptr < blksptr)
		  && (blkeptr == NIL || comptr < blkeptr)) {
		    putKcp (s, comptr-1, FALSE);
		    (void)printf("\\C{}");
		    s = comendptr;
		    putKcp (comptr, comendptr-1, FALSE);
		    incomm = TRUE;
		    comtype = STANDARD;
		    continue;
		}

	    /* start of an alternate-form comment? */
	    if (acmptr != NIL
		  && (strptr  == NIL || acmptr < strptr)
		  && (comptr  == NIL || acmptr < comptr)
		  && (chrptr  == NIL || acmptr < chrptr)
		  && (blksptr == NIL || acmptr < blksptr)
		  && (blkeptr == NIL || acmptr < blkeptr)) {
		    putKcp (s, acmptr-1, FALSE);
		    (void)printf("\\C{}");
		    s = acmendptr;
		    putKcp (acmptr, acmendptr-1, FALSE);
		    incomm = TRUE;
		    comtype = ALTERNATE;
		    continue;
		}

	    /* start of a string? */
	    if (strptr != NIL
		  && (chrptr  == NIL || strptr < chrptr)
		  && (blksptr == NIL || strptr < blksptr)
		  && (blkeptr == NIL || strptr < blkeptr)) {
		    putKcp (s, strptr-1, FALSE);
		    (void)printf("\\S{}");
		    s = strendptr;
		    putKcp (strptr,strendptr-1, FALSE);
		    instr = TRUE;
		    continue;
		}

	    /* start of a character string? */
	    if (chrptr != NIL
		  && (blksptr == NIL || chrptr < blksptr)
		  && (blkeptr == NIL || chrptr < blkeptr)) {
		    putKcp (s, chrptr-1, FALSE);
		    (void)printf("\\S{}");
		    s = chrendptr;
		    putKcp (chrptr, chrendptr-1, FALSE);
		    inchr = TRUE;
		    continue;
		}

	    /* end of a lexical block */
	    if (blkeptr != NIL) {
		if (blksptr == NIL || blkeptr < blksptr) {
		    putKcp (s, blkeendptr - 1, FALSE);
		    s = blkeendptr;
		    blklevel--;
		    if (psptr >= 0 && plstack[psptr] >= blklevel) {

			/* end of current procedure */
			blklevel = plstack[psptr];

			/* see if we should print the last proc name */
			if (--psptr >= 0)
			    prccont = TRUE;
			else
			    psptr = -1;
		    }
		    continue;
		}
	    }

	    /* start of a lexical block */
	    if (blksptr != NIL) {
		putKcp (s, blksendptr - 1, FALSE);
		s = blksendptr;
		blklevel++;
		continue;
	    }

	/* check for end of comment */
	} else if (incomm) {
	    if ((comendptr = expmatch( s,
				       comtype==STANDARD? l_comend : l_acmend,
			               (const char**)NULL, NIL)) != NIL) {
		putKcp (s, comendptr-1, TRUE);
		s = comendptr;
		incomm = FALSE;
		(void)printf("\\CE{}");
	    } else {
		comptr = s;
		s += strlen(s);
		putKcp (comptr, s-1, TRUE);
	    }
	    continue;

	/* check for end of string */
	} else if (instr) {
	    if ((strendptr =
		 expmatch (s, l_strend, (const char**)NULL, NIL)) != NIL) {
		putKcp (s, strendptr-1, TRUE);
		s = strendptr;
		instr = FALSE;
		(void)printf("\\SE{}");
	    } else {
		strptr = s;
		s += strlen(s);
		putKcp (strptr, s-1, TRUE);
	    }
	    continue;

	/* check for end of character string */
	} else if (inchr) {
	    if ((chrendptr =
		 expmatch (s, l_chrend, (const char**)NULL, NIL)) != NIL) {
		putKcp (s, chrendptr-1, TRUE);
		s = chrendptr;
		inchr = FALSE;
		(void)printf("\\SE{}");
	    } else {
		chrptr = s;
		s += strlen(s);
		putKcp (chrptr, s-1, TRUE);
	    }
	    continue;
	}

	/* print out the line */
	chrptr = s;
	s += strlen(s);
	putKcp (chrptr, s-1, FALSE);

    } while (*s);
}

#if STDC
static void
putKcp(register const char *start, register const char *end, register boolean nix)
	                     	/* start of string to write */
	                   	/* end of string to write */
	                     	/* true if we should force nokeyw */
#else
static void
putKcp(start, end, nix)
	register const char *start;	/* start of string to write */
	register const char *end;	/* end of string to write */
	register boolean nix;	/* true if we should force nokeyw */
#endif
{
	register int i, c;

	if (nokeyw)
		nix = TRUE;

	while (start <= end) {
		c = *start++;
		/* take care of nice tab stops */
		if (c == '\t') {
			while (start <= end && *start == '\t')
				start++;
			tabto(width(Str_Start, start));
			continue;
		}
#if 0
		if (!nix && (c == '#' || isinichr(c)))
#else
		/* The old test (c == '#') to handle C preprocessor lines
		   is no longer needed, since we can use :ic=\#: in C and
		   C++ language descriptions */
		if (!nix && isinichr(c)) /* then c validly begins a keyword */
#endif
		{
			/* We have a keyword if any of three conditions holds:
			   (1) we are at beginning of line
			   (2) preceding character is not an initial or
			       non-initial identifier character
			   (3) current character is not a non-initial identifier
			       character (e.g. TeX string "\global\long\def"
			       contains 3 keywords).
			   Testing (1) first ensures that start[-1] in (2) is
			   in bounds. */

			start--;	/* backup to point at initial keyword char */
			if ( (start == Str_Start)
			    || !(isidchr(start[-1]) || isinichr(start[-1]))
			    || !isidchr(c) )
			{
				i = iskw(start);
				if (i > 0) {
					(void)printf("\\K{");
					while (--i >= 0) {
						c = *start++;
						outchar(c);
					}
					putchar('}');
					continue;
				}
			}
			start++;
		}
		outchar(c);
	}
}


#if STDC
static int
width(register const char *s, register const char *os)
#else
static int
width(s, os)
	register const char *s, *os;
#endif
{
	register int i = 0, c;

	while (s < os) {
		c = *s++;
		if (c == '\t') {
			i = (i + 8) &~ 7;
			continue;
		}
		if (c < ' ')
			i += 2;
		else
			i++;
	}
	return (i);
}

/* output a string, escaping special characters */
#if STDC
static void
putstr(register const char *cp)
#else
static void
putstr(cp)
	register const char *cp;
#endif
{
	register int c;

	if (cp == NULL)
		return;
	while ((c = *cp++) != 0)
		outchar(c);
}

/*
 *	look for a process beginning on this line
 */
#if STDC
static boolean
isproc(const char *s)
#else
static boolean
isproc(s)
	const char *s;
#endif
{
	pname[0] = NUL;
	if ((!l_toplex || blklevel == 0)
	    && expmatch(s, l_prcbeg, (const char**)NULL, pname) != NIL)
		return (TRUE);
	return (FALSE);
}


/*  iskw -	check to see if the next word is a keyword
 */
#if STDC
static int
iskw(register const char *s)
#else
static int
iskw(s)
	register const char *s;
#endif
{
	register int i = 1;
	register const char *cp = s;
#if defined(LINEAR_KEYWORD_SEARCH)
	register const char **ss = (const char **)l_keywds;
	register int firstc = *s;
#endif

	while (++cp, isidchr(*cp))
		i++;
#if defined(LINEAR_KEYWORD_SEARCH)
	/* Old inefficient linear search */
	while (cp = *ss++) {
		if (!l_onecase && firstc != *cp)
			continue;
		if ((*re_strncmp)(s, cp, i) == 0 && !isidchr(cp[i]))
			return (i);
	}
	return (0);
#else
	/* New fast hash table lookup */
	return (Hash_lookup(s,i) ? i : 0);
#endif
}

#if !defined(LINEAR_KEYWORD_SEARCH)
#if STDC
static unsigned int
Hash(register const char *key, register int len)
#else
static unsigned int
Hash(key, len)
register const char *key;
register int len;
#endif
{
    register unsigned int hash_code;

    /* This hash function gives a reasonably even distribution when
       tested against the UNIX spelling dictionary, /usr/dict/words. */
    if (re_strncmp == strncmp)	/* case sensitive */
    {
	for (hash_code = 0; len > 0; ++key, --len)
	    hash_code = (hash_code << 1) ^ CHRVAL(*key);
    }
    else				/* case insensitive */
    {
	for (hash_code = 0; len > 0; ++key, --len)
	    hash_code = (hash_code << 1) ^
		CHRVAL(islower(*key) ? (int)*key : tolower(*key));
    }
    return (hash_code % MAXHASH);
}

#if STDC
static void
Hash_install(const char *key, int len)
#else
static void
Hash_install(key, len)
const char *key;
int len;
#endif
{
    register unsigned int hash_code;

    hash_code = Hash(key, len);
    while (l_keywds[hash_code] != (const char*)NULL)
    {
	if ((*re_strncmp)(key, l_keywds[hash_code], len) == 0)
	    return;			/* keyword already installed */
	hash_code++;			/* use linear probing for collisions */
	if (hash_code >= MAXHASH)
	    hash_code = 0;
    }
    Hash_used++;
    if (Hash_used > MAXHASH)
    {
	(void)fprintf(stderr, "Too many keywords for static hash table\n");
	exit (EXIT_FAILURE);
    }
    l_keywds[hash_code] = key;
}


#if STDC
static boolean
Hash_lookup(const char *key, int len)
#else
static boolean
Hash_lookup(key,len)
const char *key;
int len;
#endif
{
    register unsigned int hash_code;

    hash_code = Hash(key, len);
    while (l_keywds[hash_code] != (const char*)NULL)
    {
	if ((*re_strncmp)(key, l_keywds[hash_code], len) == 0)
	    return (TRUE);		/* key present */
	hash_code++;			/* use linear probing for collisions */
	if (hash_code >= MAXHASH)
	    hash_code = 0;
    }
    return (FALSE);			/* key absent */
}
#endif
